import ipyvuetify as v

from sepal_ui import sepalwidgets as sw
from sepal_ui import mapping as sm
from sepal_ui.message import ms

from sepal_ui.aoi.aoi_view import AoiView

__all__ = ["AoiTile"]


class AoiTile(sw.Tile):
    """
    sw.Tile tailored for the selection of an aoi. it is meant to be used with the aoi. it can be bound to EarthEngine (GAUl 2015 administrative definition) or stay with the GADM Python implementation

    Args:
        methods (list, optional): the methods to select the aoi (more information in AoiView), default to 'ALL'. Available: {‘ADMIN0’, ‘ADMIN1’, ‘ADMIN2’, ‘SHAPE’, ‘DRAW’, ‘POINTS’, ‘ASSET’, ‘ALL’}
        ee (bool, optional): wether or not to use the python EartEngine API. default to True
        vector (str|pathlib.Path, optional): the path to the default vector object
        admin (int, optional): the administrative code of the default selection. Need to be GADM if ee==False and GAUL 2015 if ee==True.
        asset (str, optional): the default asset. Can only work if :code:`ee==True`.
    """

    map = None
    "sepal_ui.mapping.SepalMap: a SepalMap object to display the selected aoi"

    view = None
    "widget: an AoiView object to handle the aoi method selection"

    def __init__(self, methods="ALL", gee=True, **kwargs):

        # create the map
        self.map = sm.SepalMap(dc=True, gee=gee)

        # create the view
        # the view include the model
        self.view = AoiView(methods=methods, map_=self.map, gee=gee, **kwargs)
        self.view.elevation = 0

        # organise them in a layout
        layout = v.Layout(
            row=True,
            xs12=True,
            children=[
                v.Flex(xs12=True, md6=True, class_="pa-5", children=[self.view]),
                v.Flex(xs12=True, md6=True, class_="pa-1", children=[self.map]),
            ],
        )

        # create the tile
        super().__init__("aoi_tile", ms.aoi_sel.title, inputs=[layout], **kwargs)



__author__ = """Pierrick Rambaud"""
__email__ = "pierrick.rambaud49@gmail.com"
__version__ = "2.5.4"

# direct access to colors
from sepal_ui.frontend import styles
import ipyvuetify as v
from types import SimpleNamespace

theme = v.theme.themes.dark if v.theme.dark else v.theme.themes.light
"traitlets: the theme used in sepal"

color = SimpleNamespace(
    bg=styles.bg_color,
    primary=theme.primary,
    accent=theme.accent,
    secondary=theme.secondary,
    success=theme.success,
    info=theme.info,
    warning=theme.warning,
    error=theme.error,
)
'SimpleNamespace: the colors of sepal. members are in the following list: "bg, primary, accent, secondary, success, info, warning, error". They will render according to the selected theme.'



from traitlets import Any, Dict, Bool, Int
from pathlib import Path
from natsort import natsorted

import pandas as pd
import geopandas as gpd
import numpy as np
import rasterio as rio
from rasterio.windows import from_bounds
from matplotlib.colors import to_rgba

from .parameters import NO_VALUE
from sepal_ui.model import Model
from sepal_ui.scripts import gee
from sepal_ui.scripts import utils as su
from sepal_ui.message import ms

import ee

__all__ = ["ReclassifyModel"]


class ReclassifyModel(Model):
    """
    Reclassification model to store information about the current reclassification and share them within your app. save all the input and output of the reclassification + the the matrix to move from one to another. It is embeding 2 backends, one based on GEE that will use assets as in/out and another based on python that will use local files as in/out. The model can handle both vector and raster data, the format and name of the output will be determined from the the input format/name. The developer will still have the possiblity to choose where to save the outputs (folder name).

    Args:
        gee (bool): either or not to set :code:`gee` to True
        dst_dir (str): the destination forlder for outputs
        folder(str, optional): the init GEE asset folder where the asset selector should start looking (debugging purpose)
        aoi_model (aoi.AoiModel, optional): the aoi model to link to the reclassify workflow
        enforce_aoi (bool, optional): either or not an aoi should be set to allow the reclassification
    """

    # inputs
    # should be unicode but we need to handle when nothing is set (None)
    band = Any(None).tag(sync=True)
    "str|int: the band name or number to use for the reclassification if raster type. Use property name if vector type"

    src_local = Any(None).tag(sync=True)
    "str: the source file to reclassify (from a local path) only used if :code:`gee=False`"

    src_gee = Any(None).tag(sync=True)
    "str: AssetId of the used input asset for reclassification. Only used if :code:`gee=True`"

    dst_class_file = Any(None).tag(sync=True)
    "str: the destination file for reclassify matrix"

    dst_dir = Any(None).tag(sync=True)
    "str: the dir used to store the output"

    gee = Bool(False).tag(sync=True)
    "bool: either to use the gee backend or not"

    aoi_model = None
    "aoi.AoiModel: AOI model object to get an area of interest if one is selected"

    folder = None
    "str: the init GEE asset folder where the asset selector should start looking (debugging purpose)"

    enforce_aoi = None
    "bool: either or not an aoi should be set to allow the reclassification"

    # data manipulation
    matrix = Dict({}).tag(sync=True)
    "dict: the transfer matrix between the input and the output using the following format: {old_value: new_value, ...}"

    # outputs
    input_type = Bool(False).tag(sync=True)  # 1 raster, 0 vector
    "bool: the input type, 1 for raster and 0 for vector"

    src_class = Dict({}).tag(sync=True)
    "dict: the source classes using the following columns: {code: (desc, color)}"

    dst_class = Dict({}).tag(sync=True)
    "dict: the destination classes using the following columns: {code: (desc, color)}"

    dst_local = Any(None).tag(sync=True)
    "str: the output file. default to :code:`dst_dir/f'{src_local.stem}_reclass.{src_local.suffix}'`"

    dst_gee = Any(None).tag(sync=True)
    "str: the output assetId. default to :code:`dst_dir/f'{src_gee.stem}_reclass'`"

    # Create a state var, to determine if an asset has been remaped
    remaped = Int(False).tag(sync=True)
    "int: state var updated each time an input is remapped"

    save = False
    "bool: either or not the relcassified dataset need to be saved"

    dst_local_memory = None
    "Any: the local output of the reclassification"

    dst_gee_memory = None
    "Any: the gee output of the reclassification"

    table_created = Bool(False).tag(sync=True)
    "bool: either or not a table have been created"

    def __init__(
        self,
        gee=False,
        dst_dir=Path.home(),
        aoi_model=None,
        folder=None,
        save=True,
        enforce_aoi=False,
        **kwargs,
    ):

        # init the model
        super().__init__(**kwargs)

        # save the folder where the results should be stored
        # only used for local export
        self.dst_dir = Path(dst_dir)

        # save relation with gee
        self.gee = gee

        if self.gee:
            su.init_ee()

        if self.gee:
            self.folder = folder or ee.data.getAssetRoots()[0]["id"]
        else:
            self.folder = None

        self.aoi_model = aoi_model

        # aoi_model and reclassify model must be aligned when it comes to gee
        if self.aoi_model:
            if self.aoi_model.ee != self.gee:
                raise Exception(
                    "Both aoi_model.gee and self.gee parameters has to be equals."
                    + f"Received {self.aoi_model.ee} for aoi_model and {self.gee} for reclassify_model."
                )

        self.enforce_aoi = enforce_aoi

        # set if the model need to save by default
        self.save = save

    def get_classes(self):
        """
        Extract the classes from the class file. The class file need to be compatible with the reclassify tool i.e. a table file with 3 headerless columns using the following format: 'code', 'desc', 'color'. Color need to be set in hexadecimal to be read else black will be used.

        Args:
            file (pathlike object): the pathlib object of the class file

        Return:
            (dict): the dict of the classes using following format:
                {code: (name, color)}
        """

        file = self.dst_class_file

        if not file:
            raise AttributeError("missing file")

        path = Path(file)
        if not path.is_file():
            raise Exception(f"{file} is not existing")

        df = pd.read_csv(file, header=None)

        # dst_class_file should be set on the model csv output of the custom view
        # 3 column: 1: code, 2: name, 3: color
        df = df.rename(columns={0: "code", 1: "desc", 2: "color"})

        # save the df for reclassify usage
        class_list = {row.code: (row.desc, row.color) for _, row in df.iterrows()}

        # create a dict out of it
        return class_list

    def get_type(self):
        """
        Guess the type of the input and set the input type attribute for the model (vector or raster)

        Return:
            (bool): the type of input (1 for raster, 0 for vector)
        """

        if self.gee:
            if not self.src_gee:
                raise Exception("Missing gee input")

            asset_info = ee.data.getAsset(self.src_gee)["type"]

            if asset_info == "TABLE":
                self.input_type = False
            elif asset_info == "IMAGE":
                self.input_type = True
            else:
                raise AttributeError(f"Unrecognized asset type: {asset_info}")
        else:
            if not self.src_local:
                raise Exception("no input")

            input_path = Path(self.src_local)

            if input_path.suffix in [".geojson", ".shp"]:
                self.input_type = False
            elif input_path.suffix in [".tif", ".tiff", ".vrt"]:
                self.input_type = True
            else:
                raise AttributeError(f"Unrecognized file format: {input_path.suffix}")
        return self.input_type

    def get_bands(self):
        """
        Use the input_type to extract all the bands/properties from the input

        Return:
            (list): sorted list of all the available bands/properties as
            integer or str
        """

        @su.need_ee
        def _ee_image():

            return ee.Image(self.src_gee).bandNames().getInfo()

        @su.need_ee
        def _ee_vector():

            columns = ee.FeatureCollection(self.src_gee).first().getInfo()["properties"]

            return (
                str(c)
                for c in columns.keys()
                if c not in ["system:index", "Shape_Area"]
            )

        def _local_image():

            with rio.open(self.src_local) as f:
                bands = [i for i in range(1, f.count + 1)]

            return bands

        def _local_vector():

            df = gpd.read_file(self.src_local)

            return [c for c in df.columns.tolist() if c != "geometry"]

        # map all the function in the guess matrix (gee, type)
        band_func = [[_local_vector, _local_image], [_ee_vector, _ee_image]]

        # return the selected function
        # remember to use self as a parameter
        return natsorted(band_func[self.gee][self.input_type]())

    def get_aoi(self):
        """Validate and get feature collection from aoi_model"""

        # return None if no aoi_model is selected
        if not self.aoi_model:
            return

        # return none if no aoi is selected
        # test gdf as it's set whatever the type of AoiModel (gee or not)
        if self.aoi_model.gdf is None:
            if self.enforce_aoi:
                raise Exception("You have to select an area of interest before")
            else:
                aoi = None

        else:  # return the aoi as a vector
            if self.gee:
                aoi = self.aoi_model.feature_collection
            else:
                aoi = self.aoi_model.gdf

        return aoi

    def unique(self):
        """
        Retreive all the existing class from the specified band/property according to the input_type.
        The data will be saved in self.src_class with no_name and black as a color.

        Return:
            (Dict): the unique class value found in the specified band/property
            and there color/name defaulted to none and black
        """

        if not self.band:
            raise Exception("You need to provide a band/property to reclassify.")

        @su.need_ee
        def _ee_image():

            # reduce the image
            image = ee.Image(self.src_gee).select(self.band)
            aoi = self.get_aoi() or image
            geometry = aoi.geometry()

            reduction = image.reduceRegion(
                ee.Reducer.frequencyHistogram(), geometry, maxPixels=1e13
            )

            # Remove all the unnecessary reducer output structure and make a
            # list of values.
            values = (
                ee.Dictionary(reduction.get(image.bandNames().get(0))).keys().getInfo()
            )

            return values

        @su.need_ee
        def _ee_vector():

            collection = ee.FeatureCollection(self.src_gee)
            aoi = self.get_aoi() or collection
            geometry = aoi.geometry()

            # get the feature
            values = (
                collection.filterBounds(geometry).aggregate_array(self.band).getInfo()
            )

            return list(set(values))

        def _local_image():

            with rio.open(self.src_local) as src:

                bounds = self.get_aoi().total_bounds if self.get_aoi() else src.bounds
                data = src.read(
                    self.band, window=from_bounds(*bounds, transform=src.transform)
                )
                count = np.bincount(data.flatten())

            return np.nonzero(count != 0)[0].tolist()

        def _local_vector():

            gdf = gpd.read_file(self.src_local)
            bounds = self.get_aoi().total_bounds if self.get_aoi() else gdf.total_bounds
            xmin, ymin, xmax, ymax = bounds
            gdf = gdf.cx[xmin:xmax, ymin:ymax]

            return gdf[self.band].unique().tolist()

        # map all the function in the guess matrix (gee, type)
        unique_func = [[_local_vector, _local_image], [_ee_vector, _ee_image]]

        # get values from the selected func
        # remember to use self as a parameter
        values = natsorted(unique_func[self.gee][self.input_type]())

        # create the init dictionnary
        self.src_class = {v: ("no_name", "#000000") for v in values}

        return self.src_class

    def reclassify(self):
        """
        Reclassify the input according to the provided matrix. For vector file type reclassifying correspond to add an extra column at the end, for raster the initial class band will be replaced by the new class, the oher being kept unmodified. vizualization colors will be set for both local (QGIS compatible) and assets (SEPAL vizualization compatible).

        Return:
            self
        """

        if not self.matrix:
            raise Exception(
                "You need a reclassification matrix to reclassify an asset."
            )
        if not self.band:
            raise Exception("You need to provide a band/property to reclassify.")

        @su.need_ee
        def _ee_image():

            if not self.src_gee:
                raise Exception("You need to provide source asset.")

            # create the asset description
            self.set_dst_gee()

            # load the image
            # remap according to the matrix
            from_, to_ = list(zip(*matrix.items()))

            image = ee.Image(self.src_gee)
            aoi = self.get_aoi() or image
            geometry = aoi.geometry()

            ee_image = (
                image.clip(geometry)
                .remap(from_, to_, 0, self.band)
                .select(["remapped"], [self.band])
            )

            # gather all the other band in the image
            # ee_image = ee.Image(self.src_gee).addBands(ee_image, overwrite=True)

            # add colormapping parameters
            # set return an element so we force cast it to ee.Image
            code, desc, color = ["0"], ["no_data"], ["#000000"]
            for k, v in self.dst_class.items():
                code.append(str(k))
                desc.append(str(v[0]))
                color.append(str(v[1]))

            ee_image = ee.Image(
                ee_image.set(
                    {
                        "visualization_0_name": "Classification",
                        "visualization_0_bands": self.band,
                        "visualization_0_type": "categorical",
                        "visualization_0_labels": ",".join(desc),
                        "visualization_0_palette": ",".join(color),
                        "visualization_0_values": ",".join(code),
                    }
                )
            )

            # save the file in a in_memeory variable
            self.dst_gee_memory = ee_image

            if self.save:

                # export
                params = {
                    "image": ee_image,
                    "assetId": self.dst_gee,
                    "description": Path(self.dst_gee).stem,
                    "scale": ee_image.projection().nominalScale().getInfo(),
                    "maxPixels": 1e13,
                    "pyramidingPolicy": {".default": "mode"},
                }

                task = ee.batch.Export.image.toAsset(**params)
                task.start()

            return self.dst_gee

        @su.need_ee
        def _ee_vector():

            if not self.src_gee:
                raise Exception("You need to provide source asset.")

            # create the asset description
            self.set_dst_gee()

            # add a new propertie

            def add_prop(feat):
                """Add reclass column to the new feature"""
                index = ee_from.indexOf(feat.get(self.band))
                # if search value is not in from, -1 is returned
                new_val = ee.Algorithms.If(index.eq(-1), NO_VALUE, ee_to.get(index))
                return feat.set({"reclass": new_val})

            ee_matrix = ee.List(list(matrix.items())).unzip()
            ee_from, ee_to = ee.List(ee_matrix.get(0)), ee.List(ee_matrix.get(1))

            collection = ee.FeatureCollection(self.src_gee)
            aoi = self.get_aoi() or collection
            geometry = aoi.geometry()

            self.dst_gee_memory = (
                ee.FeatureCollection(self.src_gee).filterBounds(geometry).map(add_prop)
            )

            # add colormapping parameters

            if self.save:

                # export
                params = {
                    "collection": self.dst_gee_memory,
                    "assetId": self.dst_gee,
                    "description": Path(self.dst_gee).stem,
                }

                task = ee.batch.Export.table.toAsset(**params)
                task.start()

            return self.dst_gee

        def _local_image():

            if not self.src_local:
                raise Exception("You need to provide source asset.")

            # set the output file
            self.dst_local = str(
                self.dst_dir / f"{Path(self.src_local).stem}_reclass.tif"
            )

            with rio.open(self.src_local) as src:

                bounds = self.get_aoi().total_bounds if self.get_aoi() else src.bounds
                window = from_bounds(*bounds, transform=src.transform)
                raw = src.read(self.band, window=window)

                data = np.zeros_like(raw, dtype=np.int64)

                for old_val, new_val in matrix.items():
                    data += (raw == old_val) * new_val

                if self.save:

                    profile = src.profile
                    profile.update(
                        driver="GTiff",
                        count=1,
                        compress="lzw",
                        dtype=np.uint8,
                        height=window.height,
                        width=window.width,
                        transform=src.window_transform(window),
                    )

                    with rio.open(self.dst_local, "w", **profile) as dst:
                        dst.write(data.astype(np.uint8), 1, window=window)

                        # add the colors to the image
                        colormap = {0: (0, 0, 0)}
                        for code, item in self.dst_class.items():
                            colormap[code] = tuple(
                                int(c * 255) for c in to_rgba(item[1])
                            )
                        dst.write_colormap(self.band, colormap)

                    # Save raster in memory
                    self.dst_local_memory = data

            return self.dst_local

        def _local_vector():

            if not self.src_local:
                raise Exception("You need need to provide source asset.")

            # set the output file
            self.dst_local = str(
                self.dst_dir / f"{Path(self.src_local).stem}_reclass.shp"
            )

            # read the dataset
            gdf = gpd.read_file(self.src_local)
            bounds = self.get_aoi().total_bounds if self.get_aoi() else gdf.total_bounds
            xmin, ymin, xmax, ymax = bounds
            gdf = gdf.cx[xmin:xmax, ymin:ymax]

            # map the new column
            gdf["reclass"] = gdf.apply(lambda row: matrix[row[self.band]], axis=1)

            # add the colors to the gdf
            # waiting for an answer there :
            # https://gis.stackexchange.com/questions/404946/how-can-i-save-my-geopandas-symbology
            self.dst_local_memory = gdf

            # save the file
            if self.save:

                gdf.to_file(self.dst_local)

            return self.dst_local

        # map all the function in the guess matrix (gee, type)
        reclassify_func = [[_local_vector, _local_image], [_ee_vector, _ee_image]]

        # reshape the matrix so that every value correspond to 1 key
        # Cast np.int64 to int
        matrix = {int(k): int(v) for k, v in self.matrix.items()}

        # return the selected function
        # remember to use self as a parameter
        res = reclassify_func[self.gee][self.input_type]()

        # tell the rest of the apps that a reclassification is finished
        self.remaped += 1

        if self.save:
            return ms.rec.rec.export[self.gee][self.input_type].format(res)

        return "Asset successfully reclassified."

    def set_dst_gee(self):
        """
        Creates a unique and consecutive asset name based on the source

        Return:
            (str) the destination folder
        """

        # create the asset_id
        asset_name = f"{Path(self.src_gee).stem}_reclass"

        # impossible to guess the folder as an asset name can start with:
        # project/earthengin-legacy/users/xx/xx/....
        # users/xx/xx/xx
        # OXFORD/xx/xx/xx for public dataset
        # fallback to the root folder
        dst_gee = str(Path(self.folder, asset_name))

        # check if the name already exist
        current_assets = [asset["name"] for asset in gee.get_assets(self.folder)]

        # An user could reclassify twice an asset,
        # So let's create an unique name
        while dst_gee in current_assets:
            dst_gee = su.next_string(dst_gee)

        # set the dst_gee of the model
        self.dst_gee = dst_gee

        return self



from IPython.display import display
import ipyvuetify as v
from traitlets import Unicode


class ResizeTrigger(v.VuetifyTemplate):
    """
    A trigger to resize maps when a change of display is done.
    Every time resize is called, the javascript resize event is trigger of the application
    """

    template = Unicode(
        """
        <script>
            {methods: {
                jupyter_resize(){
                    window.dispatchEvent(new Event('resize'));
                }
            }}
        </script>
    """
    ).tag(sync=True)
    "Unicode: the javascript script to manually trigger the resize event"

    def resize(self):
        """trigger the template method i.e. the resize event"""

        self.send({"method": "resize"})

        return


# create one single resizetrigger that will be used as a singleton everywhere
# singletons are bad but if we display multiple instances of rt for every DrawItem
# the initial offset will be impossible to manage
rt = ResizeTrigger()
display(rt)



from pathlib import Path

import ipyvuetify as v

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.sepalwidgets.widget import Markdown
from sepal_ui.scripts import utils as su

__all__ = ["Tile", "TileAbout", "TileDisclaimer"]


class Tile(v.Layout, SepalWidget):
    """
    Custom Layout widget for the sepal UI framework.
    It an helper to build a consistent tiling system.
    Tile objects are indeed compatible with the other classes from sepal_ui.

    Args:
        id_ (str): the tile id that will be written in its mount_id _metadata attribute
        title (str): the title of the Tile
        inputs ([list]): the list of widget to display inside the tile
        btn (v.Btn): the process btn
        alert (sw.Alert): the alert to display process informations to the end user
        kwargs (optional): any parameter from a v.Layout. if set, 'children' and '_metadata' will be overwritten.
    """

    btn = None
    "v.btn: the process btn"

    alert = None
    "sw.Alert: the alert to display process informations to the end user"

    title = None
    "v.Html: the title of the Tile"

    def __init__(self, id_, title, inputs=[""], btn=None, alert=None, **kwargs):

        self.btn = btn
        if btn:
            inputs.append(btn)

        self.alert = alert
        if alert:
            inputs.append(alert)

        self.title = v.Html(xs12=True, tag="h2", children=[title])

        content = [v.Flex(xs12=True, children=[widget]) for widget in inputs]

        card = v.Card(
            class_="pa-5", raised=True, xs12=True, children=[self.title] + content
        )

        # set some default parameters
        kwargs["_metadata"] = {"mount_id": id_}
        kwargs["row"] = kwargs.pop("row", True)
        kwargs["align_center"] = kwargs.pop("align_center", True)
        kwargs["class_"] = kwargs.pop("class_", "ma-5 d-inline")
        kwargs["children"] = [card]

        # call the constructor
        super().__init__(**kwargs)

    def nest(self):
        """
        Prepare the tile to be used as a nested component in a tile.
        the elevation will be set to 0 and the title remove from children.
        The mount_id will also be changed to nested

        Return:
            self
        """

        # remove id
        self._metadata["mount_id"] = "nested_tile"

        # remove elevation
        self.elevation = False

        # remove title
        self.set_title()

        return self

    def set_content(self, inputs):
        """
        Replace the current content of the tile with the provided inputs. it will keep the output and btn widget if existing.

        Args:
            inputs ([list]): the list of widget to display inside the tile

        Return:
            self
        """

        # create the widgets
        content = [v.Flex(xs12=True, children=[widget]) for widget in inputs]

        # add the output (if existing)
        if self.title:
            content = [self.title] + content

        if self.alert:
            content = content + [self.alert]

        if self.btn:
            content = content + [self.btn]

        self.children[0].children = content

        return self

    def set_title(self, title=None):
        """
        Replace the current title and activate it.
        If no title is provided, the title is removed from the tile content

        Args:
            title (str, optional): the new title of the object

        Return:
            self
        """

        # set the title text
        self.title.children = [str(title)]

        # add the title if it's deactivated
        if title and self.title not in self.children[0].children:
            self.children[0].children = [self.title] + self.children[0].children.copy()

        # remove it if it's deactivated
        elif self.title in self.children[0].children and not title:
            self.children[0].children = self.children[0].children.copy()[
                1:
            ]  # it's the first one

        return self

    def get_title(self):
        """
        Return the current title of the tile

        Return:
            (str): the title
        """

        return self.title.children[0]

    def toggle_inputs(self, fields_2_show, fields):
        """
        Display only the widgets that are part of the input_list. the widget_list is the list of all the widgets of the tile.

        Args:
            fields_2_show ([v.widget]) : the list of input to be display
            fields ([v.widget]) : the list of the tile widget

        Return:
            self
        """

        for field in fields:
            if field in fields_2_show:
                su.show_component(field)
            else:
                su.hide_component(field)

        return self

    def get_id(self):
        """
        Return the mount_id value

        Return:
            (str): the moun_id value from _metadata dict
        """

        return self._metadata["mount_id"]


class TileAbout(Tile):
    """
    Create an about tile using a .md file.
    This tile will have the "about_widget" id and "About" title.

    Args:
        pathname (str | pathlib.Path): the path to the .md file
    """

    def __init__(self, pathname, **kwargs):

        if type(pathname) == str:
            pathname = Path(pathname)

        # read the content and transform it into a html
        with pathname.open() as f:
            about = f.read()

        content = Markdown(about)

        super().__init__("about_tile", "About", inputs=[content])


class TileDisclaimer(Tile):
    """
    Create a about tile using a the generic disclaimer .md file.
    This tile will have the "about_widget" id and "Disclaimer" title.
    """

    def __init__(self):

        pathname = Path(__file__).parents[1] / "scripts" / "disclaimer.md"

        # read the content and transform it into a html
        with pathname.open() as f:
            disclaimer = f.read()

        content = Markdown(disclaimer)

        super().__init__("about_tile", "Disclaimer", inputs=[content])



import ipyvuetify as v

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from ..scripts import utils as su

__all__ = ["Btn", "DownloadBtn"]


class Btn(v.Btn, SepalWidget):
    """
    Custom process Btn filled with the provided text.
    the color will be defaulted to 'primary' and can be changed afterward according to your need

    Args:
        text (str, optional): the text to display in the btn
        icon (str, optional): the full name of any mdi-icon
        kwargs (dict, optional): any parameters from v.Btn. if set, 'children' will be overwritten.
    """

    v_icon = None
    "v.Icon: the icon in the btn"

    def __init__(self, text="Click", icon="", **kwargs):

        # create the default v_icon
        self.v_icon = v.Icon(left=True, children=[""])
        self.set_icon(icon)

        # set the default parameters
        kwargs["color"] = kwargs.pop("color", "primary")
        kwargs["children"] = [self.v_icon, text]

        # call the constructor
        super().__init__(**kwargs)

    def set_icon(self, icon=""):
        """
        set a new icon. If the icon is set to "", then it's hidden.

        Args:
            icon (str, optional): the full name of a mdi-icon

        Return:
            self
        """
        self.v_icon.children = [icon]

        if not icon:
            su.hide_component(self.v_icon)
        else:
            su.show_component(self.v_icon)

        return self

    def toggle_loading(self):
        """
        Jump between two states : disabled and loading - enabled and not loading

        Return:
            self
        """
        self.loading = not self.loading
        self.disabled = self.loading

        return self


class DownloadBtn(v.Btn, SepalWidget):
    """
    Custom download Btn filled with the provided text.
    the download icon is automatically embeded and green.
    The btn only accepts absolute links. if non is provided then the btn stays disabled

    Args:
        text (str): the message inside the btn
        path (str, optional): the absolute to a downloadable content
        args (dict, optional): any parameter from a v.Btn. if set, 'children' and 'target' will be overwritten.
    """

    def __init__(self, text, path="#", **kwargs):

        # create a download icon
        v_icon = v.Icon(left=True, children=["mdi-download"])

        # set default parameters
        kwargs["class_"] = kwargs.pop("class_", "ma-2")
        kwargs["xs5"] = kwargs.pop("xs5", True)
        kwargs["color"] = kwargs.pop("color", "success")
        kwargs["children"] = [v_icon, text]
        kwargs["target"] = "_blank"

        # call the constructor
        super().__init__(**kwargs)

        # create the URL
        self.set_url(path)

    def set_url(self, path="#"):
        """
        Set the URL of the download btn. and unable it.
        If nothing is provided the btn is disabled

        Args:
            path (str): the absolute path to a downloadable content

        Return:
            self
        """

        if su.is_absolute(path):
            url = path
        else:
            url = su.create_download_link(path)

        self.href = url
        self.disabled = path == "#"

        return self



from pathlib import Path
from traitlets import Unicode

import ipyvuetify as v
import pandas as pd

from .parameters import NO_VALUE, MATRIX_NAMES
import sepal_ui.sepalwidgets as sw
from sepal_ui.scripts import utils as su
from sepal_ui.message import ms
from sepal_ui.scripts.utils import loading_button
from .reclassify_model import ReclassifyModel

__all__ = ["ReclassifyView"]


class ImportMatrixDialog(v.Dialog):
    """
    Dialog to select the file to use and fill the matrix

    Args:
        folder (pathlike object): the path to the saved classifications

    Attributes:
        file (str): the file to use
    """

    file = Unicode("").tag(sync=True)

    def __init__(self, folder, **kwargs):

        # create the 3 widgets
        title = v.CardTitle(children=["Load reclassification matrix"])
        self.w_file = sw.FileInput(label="filename", folder=folder)
        self.load_btn = sw.Btn("Load")
        cancel = sw.Btn("Cancel", outlined=True)
        actions = v.CardActions(children=[cancel, self.load_btn])

        # default params
        self.value = False
        self.max_width = 500
        self.overlay_opacity = 0.7
        self.persistent = True
        self.children = [v.Card(class_="pa-4", children=[title, self.w_file, actions])]

        # create the dialog
        super().__init__(**kwargs)

        # js behaviour
        cancel.on_event("click", self._cancel)

    def _cancel(self, widget, event, data):
        """exit and do nothing"""

        self.value = False

        return self

    def show(self):

        self.value = True

        return self


class SaveMatrixDialog(v.Dialog):
    """
    Dialog to setup the name of the output matrix file

    Args:
        folder (pathlike object): the path to the save folder. default to ~/
    """

    def __init__(self, folder=Path.home(), **kwargs):

        # save the matrix
        self._matrix = {}
        self.folder = Path(folder)

        # create the widgets
        title = v.CardTitle(children=["Save matrix"])
        self.w_file = v.TextField(label="filename", v_model=None)
        btn = sw.Btn("Save matrix")
        cancel = sw.Btn("Cancel", outlined=True)
        actions = v.CardActions(children=[cancel, btn])
        self.alert = sw.Alert(children=["Choose a name for the output"]).show()

        # default parameters
        self.value = False
        self.max_width = 500
        self.overlay_opcity = 0.7
        self.persistent = True
        self.children = [
            v.Card(class_="pa-4", children=[title, self.w_file, self.alert, actions])
        ]

        # create the dialog
        super().__init__(**kwargs)

        # js behaviour
        cancel.on_event("click", self._cancel)
        btn.on_event("click", self._save)
        self.w_file.on_event("blur", self._sanitize)
        self.w_file.observe(self._store_info, "v_model")

    def _store_info(self, change):
        """Display where will be the file written"""

        new_val = change["new"]

        out_file = self.folder / f"{su.normalize_str(new_val)}.csv"
        msg = f"Your file will be saved as: {out_file}"

        if not new_val:
            msg = "Choose a name for the output"

        self.alert.add_msg(msg)

    def _cancel(self, widget, event, data):
        """do nothing and exit"""

        self.w_file.v_model = None
        self.value = False

        return self

    def _save(self, widget, event, data):
        """save the matrix in a specified file"""

        file = self.folder / f"{su.normalize_str(self.w_file.v_model)}.csv"

        matrix = pd.DataFrame.from_dict(self._matrix, orient="index").reset_index()
        matrix.columns = MATRIX_NAMES
        matrix.to_csv(file, index=False)

        # hide the dialog
        self.value = False

        return self

    def show(self, matrix):
        """show the dialog and set the matrix values"""

        self._matrix = matrix

        # Reset file name
        self.w_file.v_model = ""

        self.value = True

        return self

    def _sanitize(self, widget, event, data):
        """sanitize the used name when saving"""

        if not self.w_file.v_model:
            return self

        self.w_file.v_model = su.normalize_str(self.w_file.v_model)

        return self


class ClassSelect(sw.Select):
    """
    Custom widget to pick the value of a original class in the new classification system

    Args:
        new_codes(dict): the dict of the new codes to use as items {code: (name, color)}
        code (int): the orginal code of the class
    """

    def __init__(self, new_codes, old_code, **kwargs):

        # set default parameters
        self.items = [
            {"text": f"{code}: {item[0]}", "value": code}
            for code, item in new_codes.items()
        ]
        self.dense = True
        self.multiple = False
        self.chips = True
        self._metadata = {"class": old_code}
        self.v_model = None
        self.clearable = True

        # init the select
        super().__init__(**kwargs)


class ReclassifyTable(sw.SimpleTable):
    """
    Table to store the reclassifying information.
    2 columns are integrated, the new class value and the values in the original input
    One can select multiple class to be reclassify in the new classification

    Args:
        model (ReclassifyModel): model embeding the traitlet dict to store the reclassifying matrix. keys: class value in dst, values: list of values in src.
        dst_classes (dict|optional): a dictionnary that represent the classes of new the new classification table as {class_code: (class_name, class_color)}. class_code must be ints and class_name str.
        src_classes (dict|optional): the list of existing values within the input file {class_code: (class_name, class_color)}

    Attributes:
        HEADER (list): name of the column header (from, to)
        model (ReclassifyModel): the reclassifyModel object to manipulate the
            input file and save parameters
    """

    HEADERS = ms.rec.rec.headers

    def __init__(self, model, dst_classes={}, src_classes={}, **kwargs):

        # default parameters
        self.dense = True

        # create the table
        super().__init__(**kwargs)

        # save the model
        self.model = model

        # create the table elements
        self._header = [
            v.Html(
                tag="tr",
                children=[v.Html(tag="th", children=[h]) for h in self.HEADERS],
            )
        ]
        self.set_table(dst_classes, src_classes)

    def set_table(self, dst_classes, src_classes):
        """
        Rebuild the table content based on the new_classes and codes provided

        Args:
            dst_classes (dict|optional): a dictionnary that represent the classes of new the new classification table as {class_code: (class_name, class_color)}. class_code must be ints and class_name str.
            src_classes (dict|optional): the list of existing values within the input file {class_code: (class_name, class_color)}

        Return:
            self
        """

        # reset the matrix
        self.model.matrix = {code: 0 for code in src_classes.keys()}

        # create the select list
        # they need to observe each other to adapt the available class list dynamically
        self.class_select_list = {
            k: ClassSelect(dst_classes, k) for k in src_classes.keys()
        }

        rows = [
            v.Html(
                tag="tr",
                children=[
                    v.Html(tag="td", children=[f"{code}: {item[0]}"]),
                    v.Html(tag="td", children=[self.class_select_list[code]]),
                ],
            )
            for code, item in src_classes.items()
        ]

        # add an empty row at the end to make the table more visible when it's empty
        rows += [
            v.Html(
                tag="tr",
                children=[
                    v.Html(tag="td", children=[""]),
                    v.Html(
                        tag="td",
                        children=["" if len(dst_classes) else "No data available"],
                    ),
                ],
            )
        ]

        self.children = [v.Html(tag="tbody", children=self._header + rows)]

        # js behaviour
        [
            w.observe(self._update_matrix_values, "v_model")
            for w in self.class_select_list.values()
        ]

        return self

    def _update_matrix_values(self, change):
        """Update the appropriate matrix value when a Combo select change"""

        # get the code of the class in the src classification
        code = change["owner"]._metadata["class"]

        # bind it to classes in the dst classification
        self.model.matrix[code] = change["new"] if change["new"] else 0

        return self


class ReclassifyView(sw.Card):
    """
    Stand-alone Card object allowing the user to reclassify a input file. the input can be of any type (vector or raster) and from any source (local or GEE).
    The user need to provide a destination classification file (table) in the following format : 3 headless columns: 'code', 'desc', 'color'. Once all the old class have been attributed to their new class the file can be exported in the source format to local memory or GEE. the output is also savec in memory for further use in the app. It can be used as a tile in a sepal_ui app. The id_ of the tile is set to "reclassify_tile"

    Args:
        model (ReclassifyModel): the reclassify model to manipulate the
            classification dataset. default to a new one
        class_path (str,optional): Folder path containing already existing
            classes. Default to ~/
        out_path (str,optional): the folder to save the created classifications.
            default to ~/downloads
        gee (bool): either or not to set :code:`gee` to True. default to False
        dst_class (str|pathlib.Path, optional): the file to be used as destination classification. for app that require specific code system the file can be set prior and the user won't have the oportunity to change it
        default_class (dict|optional): the default classification system to use, need to point to existing sytem: {name: absolute_path}
        folder(str, optional): the init GEE asset folder where the asset selector should start looking (debugging purpose)
        save (bool, optional): Whether to write/export the result or not.
        enforce_aoi (bool, optional): either or not an aoi should be set to allow the reclassification
    """

    MAX_CLASS = 20
    "int: the number of line in the table to trigger the display of an extra toolbar and alert"

    model = None
    "ReclassifyModel: the reclassify model to manipulate the classification dataset"

    gee = None
    "bool: either being linked to gee or not (use local file or GEE asset for the rest of the app)"

    alert = None
    "sw.Alert: the alert to display informations about computation"

    title = None
    "v.Cardtitle: the title of the card"

    w_asset = None
    "sw.AssetSelect: the widget to select an asset input"

    w_raster = None
    "sw.FileInput: the widget to select a file input"

    w_image = None
    "Widget: wraper of the input. linked to w_asset if gee=True, else to w_raster"

    w_code = None
    "int|str: widget to select the band/property used as init classification in the input file"

    get_table_btn = None
    "sw.Btn: the btn to load the data in the reclassification table"

    w_dst_class_file = None
    "sw.FileInput: widget to select the new classification system file (3 headless columns: 'code', 'desc', 'color')"

    reclassify_table = None
    "ReclassifyTable: the reclassification table populated via the previous widgets"

    reclassify_btn = None
    "sw.Btn: the btn to launch the reclassifying process"

    def __init__(
        self,
        model=None,
        class_path=Path.home(),
        out_path=Path.home() / "downloads",
        gee=False,
        dst_class=None,
        default_class={},
        aoi_model=None,
        save=True,
        folder=None,
        enforce_aoi=False,
        **kwargs,
    ):

        # create metadata to make it compatible with the framwork app system
        self._metadata = {"mount_id": "reclassify_tile"}

        # init card parameters
        self.class_ = "pa-5"

        # create the object
        super().__init__(**kwargs)

        # set up a default model
        self.model = model or ReclassifyModel(
            gee=gee,
            dst_dir=out_path,
            aoi_model=aoi_model,
            folder=folder,
            save=save,
            enforce_aoi=enforce_aoi,
        )

        if enforce_aoi != self.model.enforce_aoi:
            raise Exception(
                "Both reclassify_model.gee and reclassify_view parameters has to be equals."
                + f"Received {enforce_aoi} for reclassify_view and {self.model.enforce_aoi} for reclassify_model."
            )

        # set the folders
        self.class_path = Path(class_path)
        self.out_path = Path(out_path)

        # save the gee binding
        self.gee = gee
        if gee:
            su.init_ee()

        # create an alert to display information to the user
        self.alert = sw.Alert()

        # set the title of the card
        self.title = v.CardTitle(
            children=[v.Html(tag="h2", children=[ms.rec.rec.title])]
        )

        # create the input widgets
        self.w_input_title = v.Html(
            tag="h2", children=[ms.rec.rec.input.title], class_="mt-5"
        )

        if self.gee:
            self.w_image = sw.AssetSelect(label=ms.rec.rec.input.asset, folder=folder)
        else:
            self.w_image = sw.FileInput(
                [".tif", ".vrt", ".tiff", ".geojson", ".shp"],
                label=ms.rec.rec.input.file,
            )

        self.w_code = v.Select(
            label=ms.rec.rec.input.band.label,
            hint=ms.rec.rec.input.band.hint,
            v_model=None,
            items=[],
            persistent_hint=True,
        )

        w_optional_title = v.Html(tag="h3", children=[ms.rec.rec.input.optional])
        self.w_src_class_file = sw.FileInput(
            [".csv"], label=ms.rec.rec.input.classif.label, folder=self.class_path
        )
        self.w_optional = v.ExpansionPanels(
            class_="mt-5",
            children=[
                v.ExpansionPanel(
                    children=[
                        v.ExpansionPanelHeader(children=[w_optional_title]),
                        v.ExpansionPanelContent(children=[self.w_src_class_file]),
                    ]
                )
            ],
        )

        # create the destination class widgetss
        self.w_class_title = v.Html(
            tag="h2", children=[ms.rec.rec.input.classif.title], class_="mt-5"
        )

        if not dst_class:
            self.w_dst_class_file = sw.FileInput(
                [".csv"], label=ms.rec.rec.input.classif.label, folder=self.class_path
            ).hide()
        else:
            # We could save a little of time without creating this
            self.w_dst_class_file.select_file(dst_class).hide()

        self.btn_list = [
            sw.Btn(
                "Custom",
                _metadata={"path": "custom"},
                small=True,
                class_="mr-2",
                outlined=True,
            )
        ] + [
            sw.Btn(
                f"use {name}",
                _metadata={"path": path},
                small=True,
                class_="mr-2",
                outlined=True,
            )
            for name, path in default_class.items()
        ]
        self.w_default = v.Flex(class_="mt-5", children=self.btn_list)

        # set the table and its toolbar
        self.w_table_title = v.Html(
            tag="h2", children=[ms.rec.rec.table], class_="mt-5"
        )

        self.save_dialog = SaveMatrixDialog(folder=out_path)
        self.import_dialog = ImportMatrixDialog(folder=out_path)
        self.get_table = sw.Btn(
            ms.rec.rec.input.btn, "mdi-table", color="success", small=True
        )
        self.import_table = sw.Btn(
            "import", "mdi-download", color="secondary", small=True, class_="ml-2 mr-2"
        )
        self.save_table = sw.Btn(
            "save", "mdi-content-save", color="secondary", small=True
        )
        self.reclassify_btn = sw.Btn(
            ms.rec.rec.btn, "mdi-checkerboard", small=True, disabled=True
        )

        self.toolbar = v.Toolbar(
            class_="d-flex mb-6",
            flat=True,
            children=[
                self.save_dialog,
                self.import_dialog,
                v.ToolbarTitle(children=["Actions"]),
                v.Divider(class_="mx-4", inset=True, vertical=True),
                self.get_table,
                v.Divider(class_="mx-4", inset=True, vertical=True),
                v.Flex(class_="ml-auto", children=[self.import_table, self.save_table]),
                v.Divider(class_="mx-4", inset=True, vertical=True),
                self.reclassify_btn,
            ],
        )

        self.reclassify_table = ReclassifyTable(self.model)

        # create a duplicate layout that include the alert and the different btns
        # it will be displayed if the number of class > MAX_CLASS
        self.duplicate_layout = v.Layout(
            class_="d-none", children=[self.toolbar, self.alert]
        )

        # bind to the model
        # bind to the 2 raster and asset as they cannot be displayed at the same time
        self.model = (
            self.model.bind(self.w_image, "src_gee" if self.gee else "src_local")
            .bind(self.w_code, "band")
            .bind(self.w_dst_class_file, "dst_class_file")
        )

        # create the layout
        self.children = [
            self.title,
            self.w_input_title,
            self.w_image,
            self.w_code,
            self.w_optional,
            self.w_class_title,
            self.w_default,
            self.w_dst_class_file,
            self.alert,
            self.w_table_title,
            self.toolbar,
            self.reclassify_table,
            self.duplicate_layout,
        ]

        # Decorate functions
        self.reclassify = loading_button(self.alert, self.reclassify_btn, debug=True)(
            self.reclassify
        )
        self.get_reclassify_table = loading_button(
            self.alert, self.get_table, debug=True
        )(self.get_reclassify_table)
        self.load_matrix_content = loading_button(
            self.alert, self.import_table, debug=True
        )(self.load_matrix_content)

        # JS Events
        self.import_table.on_event("click", lambda *args: self.import_dialog.show())
        self.import_dialog.load_btn.on_event("click", self.load_matrix_content)
        self.save_table.on_event(
            "click", lambda *args: self.save_dialog.show(self.model.matrix)
        )
        self.w_image.observe(self._update_band, "v_model")
        self.get_table.on_event("click", self.get_reclassify_table)
        self.reclassify_btn.on_event("click", self.reclassify)
        [btn.on_event("click", self._set_dst_class_file) for btn in self.btn_list]

    def _set_dst_class_file(self, widget, event, data):
        """Set the destination classification according to the one selected with btn. alter the widgets properties to reflect this change"""

        # get the filename
        filename = widget._metadata["path"]

        if filename == "custom":
            self.w_dst_class_file.show()
        else:
            self.w_dst_class_file.hide()
            self.w_dst_class_file.select_file(filename)

        # change the visibility of the btns
        for btn in self.btn_list:
            btn.outlined = False if btn == widget else True

        return self

    def load_matrix_content(self, widget, event, data):
        """
        Load the content of the file in the matrix. The table need to be already set to perform this operation

        Return:
            self
        """
        self.import_dialog.value = False
        file = self.import_dialog.w_file.v_model

        # exit if no files are selected
        if not file:
            raise Exception("No file has been selected")

        # exit if no table is loaded
        if not self.model.table_created:
            raise Exception("You have to get the table before.")

        # load the file
        # sanity checks
        input_data = pd.read_csv(file).fillna(NO_VALUE)

        try:
            input_data.astype("int64")
        except Exception:
            raise Exception(
                "This file may contain non supported charaters for reclassification."
            )

        if len(input_data.columns) != 2:
            # Try to identify the oclumns and subset them
            if all([colname in list(input_data.columns) for colname in MATRIX_NAMES]):
                input_data = input_data[MATRIX_NAMES]
            else:
                raise Exception(
                    "This file is not a properly formatted as classification matrix"
                )

        # check that the destination values are all available
        widget = list(self.reclassify_table.class_select_list.values())[0]
        classes = [i["value"] for i in widget.items]
        if not all(v in classes for v in input_data.dst.unique()):
            raise Exception(
                "Some of the destination data are not existing in the destination dataset"
            )
        # fill the data
        for _, row in input_data.iterrows():
            src_code, dst_code = row.src, row.dst
            if str(src_code) in self.reclassify_table.class_select_list:
                self.reclassify_table.class_select_list[
                    str(src_code)
                ].v_model = dst_code

        self.import_dialog.w_file.reset()

        return self

    def reclassify(self, widget, event, data):
        """
        Reclassify the input and store it in the appropriate format.
        The input is not saved locally to avoid memory overload.

        Return:
            self
        """

        # create the output file
        msg = self.model.reclassify()

        # display a message to the user
        self.alert.add_msg(msg, "success")

        return self

    @su.switch("loading", "disabled", on_widgets=["w_code"])
    def _update_band(self, change):
        """Update the band possibility to the available bands/properties of the input"""

        # guess the file type and save it in the model
        self.model.get_type()
        # update the bands values
        self.w_code.v_model = None
        self.w_code.items = self.model.get_bands()

        return self

    @su.switch("disabled", on_widgets=["reclassify_btn"], targets=[False])
    @su.switch("table_created", on_widgets=["model"], targets=[True])
    def get_reclassify_table(self, widget, event, data):
        """
        Display a reclassify table which will lead the user to select
        a local code 'from user' to a target code based on a classes file

        Return:
            self
        """

        # get the destination classes
        self.model.dst_class = self.model.get_classes()

        # get the src_classes
        self.model.src_class = self.model.unique()

        # if the src_class_file is set overwrite src_class:
        if self.w_src_class_file.v_model:
            self.model.src_class = self.model.get_classes()

        # reset the table
        self.reclassify_table.set_table(self.model.dst_class, self.model.src_class)

        # check if the duplicate_layout need to be displayed ?
        self.duplicate_layout.class_ = "d-none"
        if len(self.reclassify_table.children[0].children) - 1 > self.MAX_CLASS:
            self.duplicate_layout.class_ = "d-block"

        return self

    def nest_tile(self):
        """
        Prepare the view to be used as a nested component in a tile.
        the elevation will be set to 0 and the title remove from children.
        The mount_id will also be changed to nested

        Return:
            self
        """

        # remove id
        self._metadata["mount_id"] = "nested_tile"

        # remove elevation
        self.elevation = False

        # remove title
        without_title = self.children.copy()
        without_title.remove(self.title)
        self.children = without_title

        return self



from traitlets import Int
from datetime import datetime as dt

import pandas as pd
import geopandas as gpd
from shapely import geometry as sg

import sepal_ui.sepalwidgets as sw
from sepal_ui.scripts import utils as su
from sepal_ui.aoi.aoi_model import AoiModel
from sepal_ui.message import ms
from sepal_ui import color as sc

CUSTOM = AoiModel.CUSTOM
ADMIN = AoiModel.ADMIN
ALL = "All"
select_methods = AoiModel.METHODS

__all__ = ["AoiView", "select_methods"]


class MethodSelect(sw.Select):
    f"""
    A method selector. It will list the available methods for this very AoiView.
    'ALL' will select all the available methods (default)
    'ADMIN' only the admin one, 'CUSTOM' only the custom one.
    'XXX' will add the selected method to the list when '-XXX' will discard it.
    You cannot mix adding and removing behaviours.

    Params:
        methods (str|[str]): a list of methods from the available list ({', '.join(select_methods.keys())})
    """

    def __init__(self, methods="ALL", gee=True):

        # create the method list
        if methods == "ALL":
            self.methods = select_methods
        elif methods == "ADMIN":
            self.methods = {
                k: v for k, v in select_methods.items() if v["type"] == ADMIN
            }
        elif methods == "CUSTOM":
            self.methods = {
                k: v for k, v in select_methods.items() if v["type"] == CUSTOM
            }
        elif type(methods) == list:

            if any(m[0] == "-" for m in methods) != all(m[0] == "-" for m in methods):
                raise Exception("You mixed adding and removing, punk")

            if methods[0][0] == "-":

                to_remove = [method[1:] for method in methods]

                # Rewrite the methods instead of mutate the class methods
                self.methods = {
                    k: v for k, v in select_methods.items() if k not in to_remove
                }

            else:
                self.methods = {k: select_methods[k] for k in methods}
        else:
            raise Exception("I don't get what you meant")

        if not gee:
            self.methods.pop("ASSET", None)

        # build the item list with header
        prev_type = None
        items = []
        for k, m in self.methods.items():
            current_type = m["type"]

            if prev_type != current_type:
                items.append({"header": current_type})
            prev_type = current_type

            items.append({"text": m["name"], "value": k})

        # create the input
        super().__init__(label=ms.aoi_sel.method, items=items, v_model="", dense=True)


class AdminField(sw.Select):
    """
    An admin level selector. It is binded to ee (GAUL 2015) or not (GADM 2021). allows to select administrative codes taking into account the administrative parent code and displaying humanly readable administrative names.

    Args:
        level (int): The administrative level of the field
        parent (AdminField): the adminField that deal with the parent admin level of the current selector. used to narrow down the possible options
        ee (bool, optional): wether to use ee or not (default to True)

    Attributes:
        gee (bool): the earthengine status
        level (int): the admin level of the current field
        parent (AdminField): the field parent object
    """

    # the file location of the database
    FILE = AoiModel.FILE
    CODE = AoiModel.CODE
    NAME = AoiModel.NAME

    def __init__(self, level, parent=None, gee=True, **kwargs):

        # save ee state
        self.ee = gee

        # get the level info
        self.level = level
        self.parent = parent

        # init an empty widget
        self.v_model = None
        self.items = []
        self.clearable = True
        self.label = ms.aoi_sel.adm[level]
        super().__init__(**kwargs)

        # add js behaviour
        if self.parent:
            self.parent.observe(self._update, "v_model")

    def show(self):
        """
        when an admin field is shown, show its parent as well

        Return:
            self
        """

        super().show()

        if self.parent:
            self.parent.show()

        return self

    def get_items(self, filter_=None):
        """
        update the item list based on the given filter

        Params:
            filter_ (str): The code of the parent v_model to filter the current results

        Return:
            self
        """

        # extract the level list
        df = (
            pd.read_csv(self.FILE[self.ee])
            .drop_duplicates(subset=self.CODE[self.ee].format(self.level))
            .sort_values(self.NAME[self.ee].format(self.level))
        )

        # filter it
        if filter_:
            df = df[df[self.CODE[self.ee].format(self.level - 1)] == filter_]

        # formatted as a item list for a select component
        self.items = [
            {
                "text": su.normalize_str(
                    r[self.NAME[self.ee].format(self.level)], folder=False
                ),
                "value": r[self.CODE[self.ee].format(self.level)],
            }
            for _, r in df.iterrows()
        ]

        return self

    def _update(self, change):
        """update the item list of the admin select"""

        # reset v_model
        self.v_model = None

        # update the items list
        if change["new"]:
            self.get_items(change["new"])

        return self


class AoiView(sw.Card):
    """
    Versatile card object to deal with the aoi selection. multiple selection method are available (see the MethodSelector object) and the widget can be fully customizable. Can also be bound to ee (ee==True) or not (ee==False)

    Args:
        methods (list, optional): the methods to use in the widget, default to 'ALL'. Available: {'ADMIN0', 'ADMIN1', 'ADMIN2', 'SHAPE', 'DRAW', 'POINTS', 'ASSET', 'ALL'}
        map_ (SepalMap, optional): link the aoi_view to a custom SepalMap to display the output, default to None
        gee (bool, optional): wether to bind to ee or not
        vector (str|pathlib.Path, optional): the path to the default vector object
        admin (int, optional): the administrative code of the default selection. Need to be GADM if :code:`ee==False` and GAUL 2015 if :code:`ee==True`.
        asset (str, optional): the default asset. Can only work if :code:`ee==True`
    """

    # ##########################################################################
    # ###                             widget parameters                      ###
    # ##########################################################################

    updated = Int(0).tag(sync=True)
    "int: traitlets triggered every time a AOI is selected"

    ee = True
    "bool: either or not he aoi_view is connected to gee"

    folder = None
    "str: the folder name used in GEE related component, mainly used for debugging"

    model = None
    "sepal_ui.aoi.AoiModel: the model to create the AOI from the selected parameters"

    # ##########################################################################
    # ###                            the embeded widgets                     ###
    # ##########################################################################

    map_ = None
    "sepal_ui.mapping.SepalMap: the map to draw the AOI"

    w_method = None
    "widget: the widget to select the method"

    components = None
    "dict: the followingwidgets used to define AOI"

    w_admin_0 = None
    "widget: the widget used to select admin level 0"

    w_admin_1 = None
    "widget: the widget used to select admin level 1"

    w_admin_2 = None
    "widget: the widget used to select admin level 2"

    w_vector = None
    "widget: the widget used to select vector shapes"

    w_points = None
    "widget: the widget used to select points files"

    w_draw = None
    "widget: the widget used to select the name of a drawn shape (only if :code:`map_ != None`)"

    w_asset = None
    "widget: the widget used to select asset name of a featureCollection (only if :code:`gee == True`)"

    btn = None
    "sw.Btn: a default btn"

    alert = None
    "sw.Alert: a alert to display message to the end user"

    def __init__(self, methods="ALL", map_=None, gee=True, folder=None, **kwargs):

        # set ee dependencie
        self.ee = gee
        if gee:
            su.init_ee()
            self.folder = folder

        # get the model
        self.model = AoiModel(sw.Alert(), gee=gee, folder=folder, **kwargs)

        # get the map if filled
        self.map_ = map_

        # create the method widget
        self.w_method = MethodSelect(methods, gee=gee)

        # add the 6 methods blocks
        self.w_admin_0 = AdminField(0, gee=gee).get_items().hide()
        self.w_admin_1 = AdminField(1, self.w_admin_0, gee=gee).hide()
        self.w_admin_2 = AdminField(2, self.w_admin_1, gee=gee).hide()
        self.w_vector = sw.VectorField(label=ms.aoi_sel.vector).hide()
        self.w_points = sw.LoadTableField(label=ms.aoi_sel.points).hide()
        if self.map_:
            self.w_draw = sw.TextField(label=ms.aoi_sel.aoi_name).hide()
        if self.ee:
            self.w_asset = sw.VectorField(
                label=ms.aoi_sel.asset, gee=True, folder=self.folder, types=["TABLE"]
            ).hide()

        # group them together with the same key as the select_method object
        self.components = {
            "ADMIN0": self.w_admin_0,
            "ADMIN1": self.w_admin_1,
            "ADMIN2": self.w_admin_2,
            "SHAPE": self.w_vector,
            "POINTS": self.w_points,
        }
        if self.map_:
            self.components["DRAW"] = self.w_draw
        if self.ee:
            self.components["ASSET"] = self.w_asset

        # use the same alert as in the model
        self.alert = self.model.alert

        # bind the widgets to the model
        (
            self.model.bind(self.w_admin_0, "admin")
            .bind(self.w_admin_1, "admin")
            .bind(self.w_admin_2, "admin")
            .bind(self.w_vector, "vector_json")
            .bind(self.w_points, "point_json")
            .bind(self.w_method, "method")
        )
        if self.map_:
            self.model.bind(self.w_draw, "name")
        if self.ee:
            self.model.bind(self.w_asset, "asset_name")

        # add a validation btn
        self.btn = sw.Btn(ms.aoi_sel.btn)

        # create the widget
        self.children = (
            [self.w_method] + [*self.components.values()] + [self.btn, self.alert]
        )

        super().__init__(**kwargs)

        # js events
        self.w_method.observe(self._activate, "v_model")  # activate widgets
        self.btn.on_event("click", self._update_aoi)  # load the informations
        if self.map_:
            self.map_.dc.on_draw(self._handle_draw)  # handle map drawing

    @su.loading_button(debug=False)
    def _update_aoi(self, widget, event, data):
        """load the object in the model & update the map (if possible)"""

        # update the model
        self.model.set_object()

        # update the map
        if self.map_:
            [self.map_.remove_layer(lr) for lr in self.map_.layers if lr.name == "aoi"]
            self.map_.zoom_bounds(self.model.total_bounds())

            if self.ee:
                self.map_.addLayer(
                    self.model.feature_collection, {"color": sc.success}, "aoi"
                )
            else:
                self.map_.add_layer(self.model.get_ipygeojson())

            self.map_.hide_dc()

        # tell the rest of the apps that the aoi have been updated
        self.updated += 1

        return self

    def reset(self):
        """clear the aoi_model from input and remove the layer from the map (if existing)"""

        # clear the map
        if self.map_:
            [self.map_.remove_layer(lr) for lr in self.map_.layers if lr.name == "aoi"]
            # self.map_.center = [0, 0]
            # self.map_.zoom = 3

        # clear the model
        self.model.clear_attributes()

        # reset the alert
        self.alert.reset()

        # reset the view of the widgets
        self.w_method.v_model = None

        return self

    def _activate(self, change):
        """activate the adapted widgets"""

        # clear and hide the alert
        self.alert.reset()

        # deactivate or activate the dc
        # clear the geo_json saved features to start from scratch
        if self.map_:
            if change["new"] == "DRAW":
                self.map_.show_dc()
                self.model.geo_json = None
            else:
                self.map_.hide_dc()

        # clear the inputs
        [w.reset() for w in self.components.values()]

        # activate the widget
        [
            w.show() if change["new"] == k else w.hide()
            for k, w in self.components.items()
        ]

        return self

    def _handle_draw(self, target, action, geo_json):
        """handle the draw on map event"""

        # update the automatic name
        if not self.w_draw.v_model:
            self.w_draw.v_model = f'Manual_aoi_{dt.now().strftime("%Y-%m-%d_%H-%M-%S")}'

        # Init the json if it's not
        if self.model.geo_json is None:
            self.model.geo_json = {"type": "FeatureCollection", "features": []}

        # polygonize circles
        if "radius" in geo_json["properties"]["style"]:
            geo_json = self.polygonize(geo_json)

        if action == "created":  # no edit as you don't know which one to change
            self.model.geo_json["features"].append(geo_json)
        elif action == "deleted":
            self.model.geo_json["features"].remove(geo_json)

        return self

    @staticmethod
    def polygonize(geo_json):
        """
        Transform a ipyleaflet circle (a point with a radius) into a GeoJson multipolygon

        Params:
            geo_json (json): the circle geojson

        Return:
            (json): the polygonised circle
        """

        # get the input
        radius = geo_json["properties"]["style"]["radius"]
        coordinates = geo_json["geometry"]["coordinates"]

        # create shapely point
        circle = (
            gpd.GeoSeries([sg.Point(coordinates)], crs=4326)
            .to_crs(3857)
            .buffer(radius)
            .to_crs(4326)
        )

        # insert it in the geo_json
        json = geo_json
        json["geometry"] = circle[0].__geo_interface__

        return json



from setuptools import setup
from setuptools.command.develop import develop
from subprocess import check_call

version = "2.5.4"

DESCRIPTION = "Wrapper for ipyvuetify widgets to unify the display of voila dashboards in SEPAL platform"
LONG_DESCRIPTION = open("README.rst").read()


class DevelopCmd(develop):
    def run(self):
        """overwrite run command to install pre-commit hooks in dev mode"""
        check_call(["pre-commit", "install", "-t", "pre-commit", "-t", "commit-msg"])
        super().run()


setup_params = {
    "name": "sepal-ui",
    "version": version,
    "license": "MIT",
    "description": DESCRIPTION,
    "long_description": LONG_DESCRIPTION,
    "long_description_content_type": "text/x-rst",
    "author": "Pierrick Rambaud",
    "author_email": "pierrick.rambaud49@gmail.com",
    "url": "https://github.com/12rambau/sepal_ui",
    "download_url": f"https://github.com/12rambau/sepal_ui/archive/v_{version}.tar.gz",
    "keywords": ["UI", "Python", "widget", "sepal"],
    "python_requires": ">=3.6.9",
    "install_requires": [
        "haversine",
        "ipyvue>=1.7.0",  # this is the version with the class manager
        "ipyvuetify",  # it will work anyway as the widgets are build on the fly
        "geemap==0.8.9",
        "earthengine-api",
        "markdown",
        "xarray_leaflet",
        "shapely",
        "geopandas",
        "pandas",
        "deepdiff",
        "colorama",
        "Deprecated",
        "Unidecode",
        "natsort",
        "pipreqs",
        "cryptography",
    ],
    "extras_require": {
        "dev": [
            "pre-commit",
        ],
        "test": [
            "coverage",
            "pytest",
        ],
        "doc": [
            "jupyter-sphinx @ git+git://github.com/jupyter/jupyter-sphinx.git",
            "pydata-sphinx-theme",
            "sphinx-notfound-page",
            "Sphinx",
            "sphinxcontrib-spelling",
            "sphinx-copybutton",
            "pandoc",
            "m2r2",
        ],
    },
    "packages": [
        "sepal_ui",
        "sepal_ui.scripts",
        "sepal_ui.frontend",
        "sepal_ui.sepalwidgets",
        "sepal_ui.aoi",
        "sepal_ui.message",
        "sepal_ui.mapping",
        "sepal_ui.translator",
        "sepal_ui.model",
        "sepal_ui.reclassify",
    ],
    "package_data": {
        "sepal_ui": [
            "scripts/*.csv",
            "scripts/*.md",
            "scripts/*.json",
            "message/*.json",
            "bin/*",
        ]
    },
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    "cmdclass": {
        "develop": DevelopCmd,
    },
}

setup(**setup_params)



# knwon bug of rasterio
import os

if "GDAL_DATA" in list(os.environ.keys()):
    del os.environ["GDAL_DATA"]
if "PROJ_LIB" in list(os.environ.keys()):
    del os.environ["PROJ_LIB"]

import collections
from pathlib import Path
from distutils.util import strtobool

import geemap
from haversine import haversine
import numpy as np
import rioxarray
import xarray_leaflet  # do not remove: plugin for rioxarray so it is never called but always used
import matplotlib.pyplot as plt
from matplotlib import colors as mpc
from matplotlib import colorbar
import ipywidgets as widgets
from ipyleaflet import (
    AttributionControl,
    DrawControl,
    LayersControl,
    LocalTileLayer,
    ScaleControl,
    WidgetControl,
    ZoomControl,
)
from traitlets import Bool, link, observe
import ipyvuetify as v
import ipyleaflet
import ee

from sepal_ui.scripts import utils as su
from sepal_ui.message import ms

__all__ = ["SepalMap"]

# call x_array leaflet at least one
# flake8 will complain as it's a pluggin (i.e. never called)
# We don't want to ignore testing F401
xarray_leaflet


class SepalMap(geemap.Map):
    """
    The SepalMap class inherits from geemap.Map. It can thus be initialized with all its parameter.
    The map will fall back to CartoDB.DarkMatter map that well fits with the rest of the sepal_ui layout.
    Numerous methods have been added in the class to help you deal with your workflow implementation.
    It can natively display raster from .tif files and files and ee objects using methods that have the same signature as the GEE JavaScripts console.

    Args:
        basemaps ['str']: the basemaps used as background in the map. If multiple selection, they will be displayed as layers.
        dc (bool, optional): wether or not the drawing control should be displayed. default to false
        vinspector (bool, optional): Add value inspector to map, useful to inspect pixel values. default to false
        ee (bool, optional): wether or not to use the ee binding. If False none of the earthengine display fonctionalities can be used. default to True
        kwargs (optional): any parameter from a geemap.Map. if set, 'ee_initialize' will be overwritten.
    """

    # ############################################################################
    # ###                              Map parameters                          ###
    # ############################################################################

    ee = True
    "bool: either the map will use geempa binding or not"

    vinspector = Bool(False).tag(sync=True)
    "bool: either or not the datainspector is available"

    loaded_rasters = {}
    "dict: the list of loaded rasters"

    dc = None
    "ipyleaflet.DrawingControl: the drawing control of the map"

    def __init__(self, basemaps=[], dc=False, vinspector=False, gee=True, **kwargs):

        self.world_copy_jump = True

        # set the default parameters
        kwargs["ee_initialize"] = False  # we do it ourselves
        kwargs["add_google_map"] = kwargs.pop("add_google_map", False)
        kwargs["center"] = kwargs.pop("center", [0, 0])
        kwargs["zoom"] = kwargs.pop("zoom", 2)

        # Init the map
        super().__init__(**kwargs)

        # init ee
        self.ee = gee
        if gee:
            su.init_ee()

        # add the basemaps
        self.clear_layers()
        if len(basemaps):
            [self.add_basemap(basemap) for basemap in set(basemaps)]
        else:
            self.add_basemap("CartoDB.DarkMatter")

        # add the base controls
        self.clear_controls()
        self.add_control(ZoomControl(position="topright"))
        self.add_control(LayersControl(position="topright"))
        self.add_control(AttributionControl(position="bottomleft"))
        self.add_control(ScaleControl(position="bottomleft", imperial=False))

        # change the prefix
        for control in self.controls:
            if type(control) == AttributionControl:
                control.prefix = "SEPAL"

        # specific drawing control
        self.set_drawing_controls(dc)

        # Add value inspector
        self.w_vinspector = widgets.Checkbox(
            value=False,
            description="Inspect values",
            indent=False,
            layout=widgets.Layout(width="18ex"),
        )

        if vinspector:
            self.add_control(
                WidgetControl(widget=self.w_vinspector, position="topright")
            )

            link((self.w_vinspector, "value"), (self, "vinspector"))

        # Create output space for raster interaction
        self.output_r = widgets.Output(layout={"border": "1px solid black"})
        self.output_control_r = WidgetControl(
            widget=self.output_r, position="bottomright"
        )
        self.add_control(self.output_control_r)

        # define interaction with rasters
        self.on_interaction(self._raster_interaction)

    @observe("vinspector")
    def _change_cursor(self, change):
        """Method to be called when vinspector trait changes"""

        if self.vinspector:
            self.default_style = {"cursor": "crosshair"}
        else:
            self.default_style = {"cursor": "grab"}

        return

    def _raster_interaction(self, **kwargs):
        """Define a behavior when ispector checked and map clicked"""

        if kwargs.get("type") == "click" and self.vinspector:
            latlon = kwargs.get("coordinates")
            self.default_style = {"cursor": "wait"}

            local_rasters = [
                lr.name for lr in self.layers if isinstance(lr, LocalTileLayer)
            ]

            if local_rasters:

                with self.output_r:
                    self.output_r.clear_output(wait=True)

                    for lr_name in local_rasters:

                        lr = self.loaded_rasters[lr_name]
                        lat, lon = latlon

                        # Verify if the selected latlon is the image bounds
                        if any(
                            [
                                lat < lr.bottom,
                                lat > lr.top,
                                lon < lr.left,
                                lon > lr.right,
                            ]
                        ):
                            print("Location out of raster bounds")
                        else:
                            # row in pixel coordinates
                            y = int(((lr.top - lat) / abs(lr.y_res)))

                            # column in pixel coordinates
                            x = int(((lon - lr.left) / abs(lr.x_res)))

                            # get height and width
                            h, w = lr.data.shape
                            value = lr.data[y][x]
                            print(f"{lr_name}")
                            print(f"Lat: {round(lat,4)}, Lon: {round(lon,4)}")
                            print(f"x:{x}, y:{y}")
                            print(f"Pixel value: {value}")
            else:
                with self.output_r:
                    self.output_r.clear_output()

            self.default_style = {"cursor": "crosshair"}

            return

    def set_drawing_controls(self, add=False):
        """
        Create a drawing control for the map.
        It will be possible to draw rectangles, circles and polygons.

        Args:
            add (bool): either to add the dc to the object attribute or not

        return:
            self
        """

        color = v.theme.themes.dark.info

        dc = DrawControl(
            edit=False,
            marker={},
            circlemarker={},
            polyline={},
            rectangle={"shapeOptions": {"color": color}},
            circle={"shapeOptions": {"color": color}},
            polygon={"shapeOptions": {"color": color}},
        )

        self.dc = dc if add else None

        return self

    def _remove_local_raster(self, local_layer):
        """
        Remove local layer from memory

        Args:
            local_layer (str | geemap.layer): The local layer to remove or its name

        Return:
            self
        """
        name = local_layer if type(local_layer) == str else local_layer.name

        if name in self.loaded_rasters.keys():
            self.loaded_rasters.pop(name)

        return self

    def remove_last_layer(self, local=False):
        """
        Remove last added layer from Map

        Args:
            local (boolean): Specify True to only remove local last layers, otherwise will remove every last layer.

        Return:
            self
        """
        if len(self.layers) > 1:

            last_layer = self.layers[-1]

            if local:
                local_rasters = [
                    lr for lr in self.layers if isinstance(lr, LocalTileLayer)
                ]
                if local_rasters:
                    last_layer = local_rasters[-1]
                    self.remove_layer(last_layer)

                    # If last layer is local_layer, remove it from memory
                    if isinstance(last_layer, LocalTileLayer):
                        self._remove_local_raster(last_layer)
            else:
                self.remove_layer(last_layer)

                # If last layer is local_layer, remove it from memory
                if isinstance(last_layer, LocalTileLayer):
                    self._remove_local_raster(last_layer)

        return self

    @su.need_ee
    def zoom_ee_object(self, ee_geometry, zoom_out=1):
        """
        Get the proper zoom to the given ee geometry.

        Args:
            ee_geometry (ee.Geometry): the geometry to zoom on
            zoom_out (int) (optional): Zoom out the bounding zoom

        Return:
            self
        """

        # center the image
        self.centerObject(ee_geometry)

        # extract bounds from ee_object
        ee_bounds = ee_geometry.bounds().coordinates()
        coords = ee_bounds.get(0).getInfo()

        # Get (x, y) of the 4 cardinal points
        bl, br, tr, tl, _ = coords

        # Get (x, y) of the 4 cardinal points
        min_lon, min_lat = bl
        max_lon, max_lat = tr

        # zoom on these bounds
        self.zoom_bounds([min_lon, min_lat, max_lon, max_lat], zoom_out)

        return self

    def zoom_bounds(self, bounds, zoom_out=1):
        """
        Adapt the zoom to the given bounds. and center the image.

        Args:
            bounds ([coordinates]): coordinates corners as minx, miny, maxx, maxy
            zoom_out (int) (optional): Zoom out the bounding zoom

        Return:
            self
        """

        minx, miny, maxx, maxy = bounds

        # Center map to the centroid of the layer(s)
        self.center = [(maxy - miny) / 2 + miny, (maxx - minx) / 2 + minx]

        tl = (minx, maxy)
        bl = (minx, miny)
        tr = (maxx, maxy)
        br = (maxx, miny)

        maxsize = max(haversine(tl, br), haversine(bl, tr))

        lg = 40075  # number of displayed km at zoom 1
        zoom = 1
        while lg > maxsize:
            zoom += 1
            lg /= 2

        if zoom_out > zoom:
            zoom_out = zoom - 1

        self.zoom = zoom - zoom_out

        return self

    def add_raster(
        self,
        image,
        bands=None,
        layer_name="Layer_" + su.random_string(),
        colormap=plt.cm.inferno,
        x_dim="x",
        y_dim="y",
        opacity=1.0,
        fit_bounds=True,
        get_base_url=lambda _: "https://sepal.io/api/sandbox/jupyter",
        colorbar_position="bottomright",
    ):
        """
        Adds a local raster dataset to the map.

        Args:
            image (str | pathlib.Path): The image file path.
            bands (int or list, optional): The image bands to use. It can be either a number (e.g., 1) or a list (e.g., [3, 2, 1]). Defaults to None.
            layer_name (str, optional): The layer name to use for the raster. Defaults to None.
            colormap (str, optional): The name of the colormap to use for the raster, such as 'gray' and 'terrain'. More can be found at https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html. Defaults to None.
            x_dim (str, optional): The x dimension. Defaults to 'x'.
            y_dim (str, optional): The y dimension. Defaults to 'y'.
            opacity (float, optional): the opacity of the layer, default 1.0.
            fit_bounds (bool, optional): Wether or not we should fit the map to the image bounds. Default to True.
            get_base_url (callable, optional): A function taking the window URL and returning the base URL to use. It's design to work in the SEPAL environment, you only need to change it if you want to work outside of our platform. See xarray-leaflet lib for more details.
            colorbar_position (str, optional): The position of the colorbar (default to "bottomright"). set to False to remove it.
        """

        if type(image) == str:
            image = Path(image)

        if not image.is_file():
            raise Exception(ms.mapping.no_image)

        # check inputs
        if layer_name in self.loaded_rasters.keys():
            layer_name = layer_name + su.random_string()

        if isinstance(colormap, str):
            colormap = plt.cm.get_cmap(name=colormap)

        da = rioxarray.open_rasterio(image, masked=True)

        # The dataset can be too big to hold in memory, so we will chunk it into smaller pieces.
        # That will also improve performances as the generation of a tile can be done in parallel using Dask.
        da = da.chunk((1000, 1000))

        # Create a named tuple with raster bounds and resolution
        local_raster = collections.namedtuple(
            "LocalRaster",
            ("name", "left", "bottom", "right", "top", "x_res", "y_res", "data"),
        )(layer_name, *da.rio.bounds(), *da.rio.resolution(), da.data[0])

        self.loaded_rasters[layer_name] = local_raster

        multi_band = False
        if len(da.band) > 1 and type(bands) != int:
            multi_band = True
            if not bands:
                bands = [3, 2, 1]
        elif len(da.band) == 1:
            bands = 1

        if multi_band:
            da = da.rio.write_nodata(0)
        else:
            da = da.rio.write_nodata(np.nan)
        da = da.sel(band=bands)

        kwargs = {
            "m": self,
            "x_dim": x_dim,
            "y_dim": y_dim,
            "fit_bounds": fit_bounds,
            "get_base_url": get_base_url,
            # 'colorbar_position': colorbar_position, # will be uncoment when the colobared version of xarray-leaflet will be released
            "rgb_dim": "band" if multi_band else None,
            "colormap": None if multi_band else colormap,
        }

        # display the layer on the map
        layer = da.leaflet.plot(**kwargs)

        layer.name = layer_name

        layer.opacity = opacity if abs(opacity) <= 1.0 else 1.0

        return

    def show_dc(self):
        """
        show the drawing control on the map

        Return:
            self
        """

        if self.dc:
            self.dc.clear()

            if self.dc not in self.controls:
                self.add_control(self.dc)

        return self

    def hide_dc(self):
        """
        hide the drawing control of the map

        Return:
            self
        """

        if self.dc:
            self.dc.clear()

            if self.dc in self.controls:
                self.remove_control(self.dc)

        return self

    def add_colorbar(
        self,
        colors,
        cmap="viridis",
        vmin=0,
        vmax=1.0,
        index=None,
        categorical=False,
        step=None,
        height="45px",
        transparent_bg=False,
        position="bottomright",
        layer_name=None,
        **kwargs,
    ):
        """Add a colorbar to the map.

        Args:
            colors (list, optional): The set of colors to be used for interpolation. Colors can be provided in the form: * tuples of RGBA ints between 0 and 255 (e.g: (255, 255, 0) or (255, 255, 0, 255)) * tuples of RGBA floats between 0. and 1. (e.g: (1.,1.,0.) or (1., 1., 0., 1.)) * HTML-like string (e.g: “#ffff00) * a color name or shortcut (e.g: “y” or “yellow”)
            cmap (str): a matplotlib colormap default to viridis
            vmin (int, optional): The minimal value for the colormap. Values lower than vmin will be bound directly to colors[0].. Defaults to 0.
            vmax (float, optional): The maximal value for the colormap. Values higher than vmax will be bound directly to colors[-1]. Defaults to 1.0.
            index (list, optional):The values corresponding to each color. It has to be sorted, and have the same length as colors. If None, a regular grid between vmin and vmax is created.. Defaults to None.
            categorical (bool, optional): Whether or not to create a categorical colormap. Defaults to False.
            step (int, optional): The step to split the LinearColormap into a StepColormap. Defaults to None.
            height (str, optional): The height of the colormap widget. Defaults to "45px".
            position (str, optional): The position for the colormap widget. Defaults to "bottomright".
            layer_name (str, optional): Layer name of the colorbar to be associated with. Defaults to None.
            kwargs (any): any other argument of the colorbar object from matplotlib
        """

        width, height = 6.0, 0.4

        alpha = 1

        if colors is not None:

            # transform colors in hex colors
            hexcodes = [su.to_colors(c) for c in colors]

            if categorical:
                cmap = mpc.ListedColormap(hexcodes)
                vals = np.linspace(vmin, vmax, cmap.N + 1)
                norm = mpc.BoundaryNorm(vals, cmap.N)

            else:
                cmap = mpc.LinearSegmentedColormap.from_list("custom", hexcodes, N=256)
                norm = mpc.Normalize(vmin=vmin, vmax=vmax)

        elif cmap is not None:

            cmap = plt.get_cmap(cmap)
            norm = mpc.Normalize(vmin=vmin, vmax=vmax)

        else:
            raise ValueError(
                'cmap keyword or "palette" key in vis_params must be provided.'
            )

        style = "dark_background" if v.theme.dark is True else "classic"

        with plt.style.context(style):
            fig, ax = plt.subplots(figsize=(width, height))
            cb = colorbar.ColorbarBase(
                ax,
                norm=norm,
                alpha=alpha,
                cmap=cmap,
                orientation="horizontal",
                **kwargs,
            )

            # cosmetic improvement
            cb.outline.set_visible(False)  # remove border of the color bar
            ax.tick_params(size=0)  # remove ticks
            fig.patch.set_alpha(0.0)  # remove bg of the fig
            ax.patch.set_alpha(0.0)  # remove bg of the ax

        if layer_name:
            cb.set_label(layer_name)

        output = widgets.Output()
        colormap_ctrl = ipyleaflet.WidgetControl(
            widget=output,
            position=position,
            transparent_bg=True,
        )
        with output:
            output.clear_output()
            plt.show()

        self.colorbar = colormap_ctrl
        if layer_name in self.ee_layer_names:
            if "colorbar" in self.ee_layer_dict[layer_name]:
                self.remove_control(self.ee_layer_dict[layer_name]["colorbar"])
            self.ee_layer_dict[layer_name]["colorbar"] = colormap_ctrl

        self.add_control(colormap_ctrl)

        return

    def addLayer(
        self,
        ee_object,
        vis_params={},
        name=None,
        shown=True,
        opacity=1.0,
        viz_name=False,
    ):
        """
        Override the addLayer method from geemap to read the guess the vizaulization parameters the same way as in SEPAL recipes.
        If the vizparams are empty and vizualization metadata exist, SepalMap will use them automatically.

        Args:
            ee_object (ee.Object): the ee OBject to draw on the map
            vis_params (dict, optional): the visualization parameters set as in GEE
            name (str, optional): the name of the layer
            shown (bool, optional): either to show the layer or not, default to true (it is bugged in ipyleaflet)
            opacity (float, optional): the opcity of the layer from 0 to 1, default to 1.
            viz_name (str, optional): the name of the vizaulization you want ot use. default to the first one if existing
        """

        # get the list of viz params
        viz = self.get_viz_params(ee_object)

        # get the requested vizparameters name
        # if non is set use the first one
        if not viz == {}:
            viz_name = viz_name or viz[next(iter(viz))]["name"]

        # apply it to vis_params
        if vis_params == {} and viz != {}:

            # find the viz params in the list
            try:
                vis_params = next(i for p, i in viz.items() if i["name"] == viz_name)
            except StopIteration:
                raise ValueError(
                    f"the provided viz_name ({viz_name}) cannot be found in the image metadata"
                )

            # invert the bands if needed
            inverted = vis_params.pop("inverted", None)
            if inverted is not None:

                # get the index of the bands taht need to be inverted
                index_list = [i for i, v in enumerate(inverted) if v is True]

                # multiply everything by -1
                for i in index_list:
                    min_ = vis_params["min"][i]
                    max_ = vis_params["max"][i]
                    vis_params["min"][i] = max_
                    vis_params["max"][i] = min_

            # specific case of categorical images
            # Pad the palette when using non-consecutive values
            # instead of remapping or using sldStyle
            # to preserve the class values in the image, for inspection
            if vis_params["type"] == "categorical":

                colors = vis_params["palette"]
                values = vis_params["values"]
                min_ = min(values)
                max_ = max(values)

                # set up a black palette of correct length
                palette = ["#000000"] * (max_ - min_ + 1)

                # replace the values within the palette
                for i, val in enumerate(values):
                    palette[val - min_] = colors[i]

                # adapt the vizparams
                vis_params["palette"] = palette
                vis_params["min"] = min_
                vis_params["max"] = max_

            # specific case of hsv
            elif vis_params["type"] == "hsv":

                # set to_min to 0 and to_max to 1
                # in the original expression:
                # 'to_min + (v - from_min) * (to_max - to_min) / (from_max - from_min)'
                expression = (
                    "{band} = (b('{band}') - {from_min}) / ({from_max} - {from_min})"
                )

                # get the maxs and mins
                # removing them from the parameter
                mins = vis_params.pop("min")
                maxs = vis_params.pop("max")

                # create the rgb bands
                asset = ee_object
                for i, band in enumerate(vis_params["bands"]):

                    # adapt the expression
                    exp = expression.format(
                        from_min=mins[i], from_max=maxs[i], band=band
                    )
                    asset = asset.addBands(asset.expression(exp), [band], True)

                # set the arguments
                ee_object = asset.select(vis_params["bands"]).hsvToRgb()
                vis_params["bands"] = ["red", "green", "blue"]

        # call the function using the replacing the empty viz params with the new one.
        super().addLayer(ee_object, vis_params, name, shown, opacity)

        return

    @staticmethod
    def get_basemap_list():
        """
        This function is intending for development use
        It give the list of all the available basemaps for SepalMap object

        Return:
            ([str]): the list of the basemap names
        """

        return [k for k in geemap.ee_basemaps.keys()]

    @staticmethod
    def get_viz_params(image):
        """
        Return the vizual parmaeters that are set in the metadata of the image

        Args:
            image (ee.Image): the image to analyse

        Return:
            (dict): the dictionnary of the find properties
        """

        # the constant prefix for SEPAL visualization parameters
        PREFIX = "visualization"

        # init the property list
        props = {}

        # check image type
        if not isinstance(image, ee.Image):
            return props

        # check that image have properties
        if "properties" not in image.getInfo():
            return props

        # build a raw prop list
        raw_prop_list = {
            p: val
            for p, val in image.getInfo()["properties"].items()
            if p.startswith(PREFIX)
        }

        # decompose each property by its number
        # and gather the properties in a sub dictionnary
        for p, val in raw_prop_list.items():

            # extract the number and create the sub-dict
            _, number, name = p.split("_")
            props[number] = props.pop(number, {})

            # modify the values according to prop key
            if isinstance(val, str):
                if name in ["bands", "palette", "labels"]:
                    val = val.split(",")
                elif name in ["max", "min", "values"]:
                    val = [float(i) for i in val.split(",")]
                elif name in ["inverted"]:
                    val = [bool(strtobool(i)) for i in val.split(",")]

            # set the value
            props[number][name] = val

        # categorical values need to be cast to int
        for i in props.keys():
            if props[i]["type"] == "categorical":
                props[i]["values"] = [int(val) for val in props[i]["values"]]

        return props



import os
from pathlib import Path

from sepal_ui.translator import Translator

# pick the language
lang = "en"
if "CUSTOM_LANGUAGE" in os.environ:
    lang = os.environ["CUSTOM_LANGUAGE"]

ms = Translator(Path(__file__).parent, lang)



import json
from types import SimpleNamespace
from pathlib import Path
from collections import abc
from deepdiff import DeepDiff


class Translator(SimpleNamespace):
    """
    The translator is a SimpleNamespace of Simplenamespace. It reads 2 Json files, the first one being the source language (usually English) and the second one the target language.
    It will replace in the source dictionary every key that exist in both json dictionaries. Following this procedure, every message that is not translated can still be accessed in the source language.
    To access the dictionary keys, instead of using [], you can simply use key name as in an object ex: translator.first_key.secondary_key.
    There are no depth limits, just respect the snake_case convention when naming your keys in the .json files.

    Args:
        json_folder (str | pathlib.Path): the folder where the dictionaries are stored
        target_lan (str): the language code of the target lang (it should be the same as the target dictionary)
        default_lan (str): the language code of the source lang (it should be the same as the source dictionary)
    """

    FORBIDDEN_KEYS = ["default_dict", "target_dict", "in", "class"]
    "list(str): list of the forbidden keys, using one of them in a translation dict will throw an error"

    target_dict = {}
    "(dict): the target language dictionary"

    default_dict = {}
    "dict: the source language dictionary"

    keys = None
    "all the keys can be acceced as attributes"

    def __init__(self, json_folder, target_lan, default_lan="en"):

        super().__init__()

        if type(json_folder) == str:
            json_folder = Path(json_folder)

        # reading both the default dict
        source_path = json_folder / f"{default_lan}.json"
        self.default_dict = json.loads(source_path.read_text())

        # create a composite dict replaceing all the default keys with the one availabel in the target lan
        target_path = json_folder / f"{target_lan}.json"
        self.target_dict = self.default_dict.copy()

        if target_path.is_file():
            self.target_dict = json.loads(target_path.read_text())
        else:
            print(f'No json file is provided for "{target_lan}", fallback to "en"')

        # create the composite dictionary
        ms_dict = self._update(self.default_dict, self.target_dict)

        # verify if 'default_dict' or 'target_dict' is in use
        [self.search_key(ms_dict, k) for k in self.FORBIDDEN_KEYS]

        # transform it into a json str
        ms_json = json.dumps(ms_dict)

        # unpack the json as a simple namespace
        ms = json.loads(ms_json, object_hook=lambda d: SimpleNamespace(**d))

        for k, v in ms.__dict__.items():
            setattr(self, k, getattr(ms, k))

    @classmethod
    def search_key(cls, d, key):
        """
        Search a specific key in the d dictionary and raise an error if found

        Args:
            d (dict): the dictionary to study
            key (str): the key to look for
        """

        for k, v in d.items():
            if isinstance(v, abc.Mapping):
                cls.search_key(v, key)
            else:
                if k == key:
                    raise Exception(
                        f"You cannot use the key {key} in your translation dictionary"
                    )

        return

    def _update(self, d, u):
        """
        Update the fallback dictionnaire (d) values with the keys that exist in the target (u) dictionnaire

        Args:
            d (dict): The fallback dictionary
            u (dict): the target dctionnary

        Return:
            ms (dict): The updated dictionnay
        """

        ms = d.copy()

        for k, v in u.items():
            if isinstance(v, abc.Mapping):
                ms[k] = self._update(d.get(k, {}), v)
            else:
                ms[k] = v

        return ms

    def missing_keys(self):
        """
        this function is intended for developer use only
        print the list of the missing keys in the target dictionnairie

        Return:
            (str): the list of missing keys
        """

        # find all the missing keys
        try:
            ddiff = DeepDiff(self.default_dict, self.target_dict)[
                "dictionary_item_removed"
            ]
        except Exception:
            ddiff = ["All messages are translated"]

        return "\n".join(ddiff)
